namespace BitBadger.Documents.Sqlite

open Microsoft.Data.Sqlite
open WithConn

/// <summary>F# extensions for the SqliteConnection type</summary>
[<AutoOpen>]
module Extensions =

    type SqliteConnection with

        /// <summary>Execute a query that returns a list of results</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function between the document and the domain item</param>
        /// <returns>A list of results for the given query</returns>
        member conn.customList<'TDoc> query parameters mapFunc =
            Custom.list<'TDoc> query parameters mapFunc conn

        /// <summary>Execute a query that returns a JSON array of results</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function to extract the document</param>
        /// <returns>A JSON array of results for the given query</returns>
        member conn.customJsonArray query parameters mapFunc =
            Custom.jsonArray query parameters mapFunc conn
        
        /// <summary>Execute a query, writing its results to the given <c>StreamWriter</c></summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="writer">The StreamWriter to which the results should be written</param>
        /// <param name="mapFunc">The mapping function to extract the document</param>
        member conn.writeCustomJsonArray query parameters writer mapFunc =
            Custom.writeJsonArray query parameters writer mapFunc conn

        /// <summary>Execute a query that returns one or no results</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function between the document and the domain item</param>
        /// <returns><c>Some</c> with the first matching result, or <c>None</c> if not found</returns>
        member conn.customSingle<'TDoc> query parameters mapFunc =
            Custom.single<'TDoc> query parameters mapFunc conn

        /// <summary>Execute a query that returns one or no JSON documents</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function to extract the document</param>
        /// <returns>The JSON document with the first matching result, or an empty document if not found</returns>
        member conn.customJsonSingle query parameters mapFunc =
            Custom.jsonSingle query parameters mapFunc conn

        /// <summary>Execute a query that returns no results</summary>
        /// <param name="query">The query to retrieve the results</param>
        /// <param name="parameters">Parameters to use for the query</param>
        member conn.customNonQuery query parameters =
            Custom.nonQuery query parameters conn

        /// <summary>Execute a query that returns a scalar value</summary>
        /// <param name="query">The query to retrieve the value</param>
        /// <param name="parameters">Parameters to use for the query</param>
        /// <param name="mapFunc">The mapping function to obtain the value</param>
        /// <returns>The scalar value for the query</returns>
        member conn.customScalar<'T when 'T: struct> query parameters mapFunc =
            Custom.scalar<'T> query parameters mapFunc conn

        /// <summary>Create a document table</summary>
        /// <param name="name">The table whose existence should be ensured (may include schema)</param>
        member conn.ensureTable name =
            Definition.ensureTable name conn

        /// <summary>Create an index on field(s) within documents in the specified table</summary>
        /// <param name="tableName">The table to be indexed (may include schema)</param>
        /// <param name="indexName">The name of the index to create</param>
        /// <param name="fields">One or more fields to be indexed</param>
        member conn.ensureFieldIndex tableName indexName fields =
            Definition.ensureFieldIndex tableName indexName fields conn

        /// <summary>Insert a new document</summary>
        /// <param name="tableName">The table into which the document should be inserted (may include schema)</param>
        /// <param name="document">The document to be inserted</param>
        member conn.insert<'TDoc> tableName (document: 'TDoc) =
            insert<'TDoc> tableName document conn

        /// <summary>
        /// Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
        /// </summary>
        /// <param name="tableName">The table into which the document should be saved (may include schema)</param>
        /// <param name="document">The document to be saved</param>
        member conn.save<'TDoc> tableName (document: 'TDoc) =
            save tableName document conn

        /// <summary>Count all documents in a table</summary>
        /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
        /// <returns>The count of the documents in the table</returns>
        member conn.countAll tableName =
            Count.all tableName conn

        /// <summary>Count matching documents using JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
        /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <returns>The count of matching documents in the table</returns>
        member conn.countByFields tableName howMatched fields =
            Count.byFields tableName howMatched fields conn

        /// <summary>Determine if a document exists for the given ID</summary>
        /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
        /// <param name="docId">The ID of the document whose existence should be checked</param>
        /// <returns>True if a document exists, false if not</returns>
        member conn.existsById tableName (docId: 'TKey) =
            Exists.byId tableName docId conn

        /// <summary>Determine if a document exists using JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
        /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <returns>True if any matching documents exist, false if not</returns>
        member conn.existsByFields tableName howMatched fields =
            Exists.byFields tableName howMatched fields conn

        /// <summary>Retrieve all documents in the given table</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <returns>All documents from the given table</returns>
        member conn.findAll<'TDoc> tableName =
            Find.all<'TDoc> tableName conn

        /// <summary>Retrieve all documents in the given table ordered by the given fields in the document</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <returns>All documents from the given table, ordered by the given fields</returns>
        member conn.findAllOrdered<'TDoc> tableName orderFields =
            Find.allOrdered<'TDoc> tableName orderFields conn

        /// <summary>Retrieve a document by its ID</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="docId">The ID of the document to retrieve</param>
        /// <returns><c>Some</c> with the document if found, <c>None</c> otherwise</returns>
        member conn.findById<'TKey, 'TDoc> tableName (docId: 'TKey) =
            Find.byId<'TKey, 'TDoc> tableName docId conn

        /// <summary>Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <returns>All documents matching the given fields</returns>
        member conn.findByFields<'TDoc> tableName howMatched fields =
            Find.byFields<'TDoc> tableName howMatched fields conn

        /// <summary>
        /// Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in
        /// the document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <returns>All documents matching the given fields, ordered by the other given fields</returns>
        member conn.findByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields =
            Find.byFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields conn

        /// <summary>Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <returns><c>Some</c> with the first document, or <c>None</c> if not found</returns>
        member conn.findFirstByFields<'TDoc> tableName howMatched fields =
            Find.firstByFields<'TDoc> tableName howMatched fields conn

        /// <summary>
        /// Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the
        /// given fields in the document
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <returns>
        /// <c>Some</c> with the first document ordered by the given fields, or <c>None</c> if not found
        /// </returns>
        member conn.findFirstByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields =
            Find.firstByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields conn

        /// <summary>Retrieve all JSON documents in the given table</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <returns>All JSON documents from the given table</returns>
        member conn.jsonAll tableName =
            Json.all tableName conn

        /// <summary>Retrieve all JSON documents in the given table ordered by the given fields in the document</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <returns>All JSON documents from the given table, ordered by the given fields</returns>
        member conn.jsonAllOrdered tableName orderFields =
            Json.allOrdered tableName orderFields conn

        /// <summary>Retrieve a JSON document by its ID</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="docId">The ID of the document to retrieve</param>
        /// <returns>The JSON document if found, an empty JSON document otherwise</returns>
        member conn.jsonById<'TKey> tableName (docId: 'TKey) =
            Json.byId tableName docId conn

        /// <summary>Retrieve JSON documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <returns>All JSON documents matching the given fields</returns>
        member conn.jsonByFields tableName howMatched fields =
            Json.byFields tableName howMatched fields conn

        /// <summary>
        /// Retrieve JSON documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given
        /// fields in the document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <returns>All JSON documents matching the given fields, ordered by the other given fields</returns>
        member conn.jsonByFieldsOrdered tableName howMatched queryFields orderFields =
            Json.byFieldsOrdered tableName howMatched queryFields orderFields conn

        /// <summary>
        /// Retrieve the first JSON document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <returns>The first JSON document if found, an empty JSON document otherwise</returns>
        member conn.jsonFirstByFields tableName howMatched fields =
            Json.firstByFields tableName howMatched fields conn

        /// <summary>
        /// Retrieve the first JSON document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the
        /// given fields in the document
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        /// <returns>The first JSON document (in order) if found, an empty JSON document otherwise</returns>
        member conn.jsonFirstByFieldsOrdered tableName howMatched queryFields orderFields =
            Json.firstByFieldsOrdered tableName howMatched queryFields orderFields conn

        /// <summary>Write all JSON documents in the given table to the given <c>StreamWriter</c></summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="writer">The StreamWriter to which the results should be written</param>
        member conn.writeJsonAll tableName writer =
            Json.writeAll tableName writer conn

        /// <summary>
        /// Write all JSON all documents in the given table to the given <c>StreamWriter</c>, ordered by the given
        /// fields in the document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="writer">The StreamWriter to which the results should be written</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        member conn.writeJsonAllOrdered tableName writer orderFields =
            Json.writeAllOrdered tableName writer orderFields conn

        /// <summary>Write a JSON document to the given <c>StreamWriter</c> by its ID</summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="writer">The StreamWriter to which the results should be written</param>
        /// <param name="docId">The ID of the document to retrieve</param>
        member conn.writeJsonById<'TKey> tableName writer (docId: 'TKey) =
            Json.writeById tableName writer docId conn

        /// <summary>
        /// Write JSON documents to the given <c>StreamWriter</c> matching JSON field comparisons (<c>-&gt;&gt; =</c>,
        /// etc.)
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="writer">The StreamWriter to which the results should be written</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        member conn.writeJsonByFields tableName writer howMatched fields =
            Json.writeByFields tableName writer howMatched fields conn

        /// <summary>
        /// Write JSON documents to the given <c>StreamWriter</c> matching JSON field comparisons (<c>-&gt;&gt; =</c>,
        /// etc.) ordered by the given fields in the document
        /// </summary>
        /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
        /// <param name="writer">The StreamWriter to which the results should be written</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        member conn.writeJsonByFieldsOrdered tableName writer howMatched queryFields orderFields =
            Json.writeByFieldsOrdered tableName writer howMatched queryFields orderFields conn

        /// <summary>
        /// Write the first JSON document to the given <c>StreamWriter</c> matching JSON field comparisons
        /// (<c>-&gt;&gt; =</c>, etc.)
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="writer">The StreamWriter to which the results should be written</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        member conn.writeJsonFirstByFields tableName writer howMatched fields =
            Json.writeFirstByFields tableName writer howMatched fields conn

        /// <summary>
        /// Write the first JSON document to the given <c>StreamWriter</c> matching JSON field comparisons
        /// (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in the document
        /// </summary>
        /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
        /// <param name="writer">The StreamWriter to which the results should be written</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="queryFields">The field conditions to match</param>
        /// <param name="orderFields">Fields by which the results should be ordered</param>
        member conn.writeJsonFirstByFieldsOrdered tableName writer howMatched queryFields orderFields =
            Json.writeFirstByFieldsOrdered tableName writer howMatched queryFields orderFields conn

        /// <summary>Update (replace) an entire document by its ID</summary>
        /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
        /// <param name="docId">The ID of the document to be updated (replaced)</param>
        /// <param name="document">The new document</param>
        member conn.updateById tableName (docId: 'TKey) (document: 'TDoc) =
            Update.byId tableName docId document conn

        /// <summary>
        /// Update (replace) an entire document by its ID, using the provided function to obtain the ID from the
        /// document
        /// </summary>
        /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
        /// <param name="idFunc">The function to obtain the ID of the document</param>
        /// <param name="document">The new document</param>
        member conn.updateByFunc tableName (idFunc: 'TDoc -> 'TKey) (document: 'TDoc) =
            Update.byFunc tableName idFunc document conn

        /// <summary>Patch a document by its ID</summary>
        /// <param name="tableName">The table in which a document should be patched (may include schema)</param>
        /// <param name="docId">The ID of the document to patch</param>
        /// <param name="patch">The partial document to patch the existing document</param>
        member conn.patchById tableName (docId: 'TKey) (patch: 'TPatch) =
            Patch.byId tableName docId patch conn

        /// <summary>
        /// Patch documents using a JSON field comparison query in the <c>WHERE</c> clause (<c>-&gt;&gt; =</c>, etc.)
        /// </summary>
        /// <param name="tableName">The table in which documents should be patched (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="patch">The partial document to patch the existing document</param>
        member conn.patchByFields tableName howMatched fields (patch: 'TPatch) =
            Patch.byFields tableName howMatched fields patch conn

        /// <summary>Remove fields from a document by the document's ID</summary>
        /// <param name="tableName">The table in which a document should be modified (may include schema)</param>
        /// <param name="docId">The ID of the document to modify</param>
        /// <param name="fieldNames">One or more field names to remove from the document</param>
        member conn.removeFieldsById tableName (docId: 'TKey) fieldNames =
            RemoveFields.byId tableName docId fieldNames conn

        /// <summary>Remove fields from documents via a comparison on JSON fields in the document</summary>
        /// <param name="tableName">The table in which documents should be modified (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        /// <param name="fieldNames">One or more field names to remove from the matching documents</param>
        member conn.removeFieldsByFields tableName howMatched fields fieldNames =
            RemoveFields.byFields tableName howMatched fields fieldNames conn

        /// <summary>Delete a document by its ID</summary>
        /// <param name="tableName">The table in which a document should be deleted (may include schema)</param>
        /// <param name="docId">The ID of the document to delete</param>
        member conn.deleteById tableName (docId: 'TKey) =
            Delete.byId tableName docId conn

        /// <summary>Delete documents by matching a JSON field comparison query (<c>-&gt;&gt; =</c>, etc.)</summary>
        /// <param name="tableName">The table in which documents should be deleted (may include schema)</param>
        /// <param name="howMatched">Whether to match any or all of the field conditions</param>
        /// <param name="fields">The field conditions to match</param>
        member conn.deleteByFields tableName howMatched fields =
            Delete.byFields tableName howMatched fields conn


open System.Runtime.CompilerServices

/// <summary>C# extensions on the SqliteConnection type</summary>
type SqliteConnectionCSharpExtensions =

    /// <summary>Execute a query that returns a list of results</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns>A list of results for the given query</returns>
    [<Extension>]
    static member inline CustomList<'TDoc>(conn, query, parameters, mapFunc: System.Func<SqliteDataReader, 'TDoc>) =
        Custom.List<'TDoc>(query, parameters, mapFunc, conn)

    /// <summary>Execute a query that returns a JSON array of results</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>A JSON array of results for the given query</returns>
    [<Extension>]
    static member inline CustomJsonArray(conn, query, parameters, mapFunc) =
        Custom.JsonArray(query, parameters, mapFunc, conn)

    /// <summary>Execute a query, writing its results to the given <c>StreamWriter</c></summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    [<Extension>]
    static member inline WriteCustomJsonArray(conn, query, parameters, writer, mapFunc) =
        Custom.WriteJsonArray(query, parameters, writer, mapFunc, conn)
        
    /// <summary>Execute a query that returns one or no results</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns>The first matching result, or <c>null</c> if not found</returns>
    [<Extension>]
    static member inline CustomSingle<'TDoc when 'TDoc: null and 'TDoc: not struct>(
            conn, query, parameters, mapFunc: System.Func<SqliteDataReader, 'TDoc>) =
        Custom.Single<'TDoc>(query, parameters, mapFunc, conn)

    /// <summary>Execute a query that returns one or no JSON documents</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>The JSON document with the first matching result, or an empty document if not found</returns>
    [<Extension>]
    static member inline CustomJsonSingle(conn, query, parameters, mapFunc) =
        Custom.JsonSingle(query, parameters, mapFunc, conn)

    /// <summary>Execute a query that returns no results</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    [<Extension>]
    static member inline CustomNonQuery(conn, query, parameters) =
        Custom.nonQuery query parameters conn

    /// <summary>Execute a query that returns a scalar value</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="query">The query to retrieve the value</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to obtain the value</param>
    /// <returns>The scalar value for the query</returns>
    [<Extension>]
    static member inline CustomScalar<'T when 'T: struct>(
            conn, query, parameters, mapFunc: System.Func<SqliteDataReader, 'T>) =
        Custom.Scalar<'T>(query, parameters, mapFunc, conn)

    /// <summary>Create a document table</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="name">The table whose existence should be ensured (may include schema)</param>
    [<Extension>]
    static member inline EnsureTable(conn, name) =
        Definition.ensureTable name conn

    /// <summary>Create an index on field(s) within documents in the specified table</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table to be indexed (may include schema)</param>
    /// <param name="indexName">The name of the index to create</param>
    /// <param name="fields">One or more fields to be indexed</param>
    [<Extension>]
    static member inline EnsureFieldIndex(conn, tableName, indexName, fields) =
        Definition.ensureFieldIndex tableName indexName fields conn

    /// <summary>Insert a new document</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table into which the document should be inserted (may include schema)</param>
    /// <param name="document">The document to be inserted</param>
    [<Extension>]
    static member inline Insert<'TDoc>(conn, tableName, document: 'TDoc) =
        insert<'TDoc> tableName document conn

    /// <summary>Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table into which the document should be saved (may include schema)</param>
    /// <param name="document">The document to be saved</param>
    [<Extension>]
    static member inline Save<'TDoc>(conn, tableName, document: 'TDoc) =
        save<'TDoc> tableName document conn

    /// <summary>Count all documents in a table</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
    /// <returns>The count of the documents in the table</returns>
    [<Extension>]
    static member inline CountAll(conn, tableName) =
        Count.all tableName conn

    /// <summary>Count matching documents using JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>The count of matching documents in the table</returns>
    [<Extension>]
    static member inline CountByFields(conn, tableName, howMatched, fields) =
        Count.byFields tableName howMatched fields conn

    /// <summary>Determine if a document exists for the given ID</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
    /// <param name="docId">The ID of the document whose existence should be checked</param>
    /// <returns>True if a document exists, false if not</returns>
    [<Extension>]
    static member inline ExistsById<'TKey>(conn, tableName, docId: 'TKey) =
        Exists.byId tableName docId conn

    /// <summary>Determine if a document exists using JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>True if any matching documents exist, false if not</returns>
    [<Extension>]
    static member inline ExistsByFields(conn, tableName, howMatched, fields) =
        Exists.byFields tableName howMatched fields conn

    /// <summary>Retrieve all documents in the given table</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <returns>All documents from the given table</returns>
    [<Extension>]
    static member inline FindAll<'TDoc>(conn, tableName) =
        Find.All<'TDoc>(tableName, conn)

    /// <summary>Retrieve all documents in the given table ordered by the given fields in the document</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents from the given table, ordered by the given fields</returns>
    [<Extension>]
    static member inline FindAllOrdered<'TDoc>(conn, tableName, orderFields) =
        Find.AllOrdered<'TDoc>(tableName, orderFields, conn)

    /// <summary>Retrieve a document by its ID</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    /// <returns>The document if found, <c>null</c> otherwise</returns>
    [<Extension>]
    static member inline FindById<'TKey, 'TDoc when 'TDoc: null and 'TDoc: not struct>(conn, tableName, docId: 'TKey) =
        Find.ById<'TKey, 'TDoc>(tableName, docId, conn)

    /// <summary>Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>All documents matching the given fields</returns>
    [<Extension>]
    static member inline FindByFields<'TDoc>(conn, tableName, howMatched, fields) =
        Find.ByFields<'TDoc>(tableName, howMatched, fields, conn)

    /// <summary>
    /// Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in
    /// the document
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given fields, ordered by the other given fields</returns>
    [<Extension>]
    static member inline FindByFieldsOrdered<'TDoc>(conn, tableName, howMatched, queryFields, orderFields) =
        Find.ByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields, conn)

    /// <summary>Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>The first document, or <c>null</c> if not found</returns>
    [<Extension>]
    static member inline FindFirstByFields<'TDoc when 'TDoc: null and 'TDoc: not struct>(
            conn, tableName, howMatched, fields) =
        Find.FirstByFields<'TDoc>(tableName, howMatched, fields, conn)

    /// <summary>
    /// Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given
    /// fields in the document
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first document ordered by the given fields, or <c>null</c> if not found</returns>
    [<Extension>]
    static member inline FindFirstByFieldsOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(
            conn, tableName, howMatched, queryFields, orderFields) =
        Find.FirstByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields, conn)

    /// <summary>Retrieve all JSON documents in the given table</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <returns>All JSON documents from the given table</returns>
    [<Extension>]
    static member inline JsonAll(conn, tableName) =
        Json.all tableName conn

    /// <summary>Retrieve all JSON documents in the given table ordered by the given fields in the document</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All JSON documents from the given table, ordered by the given fields</returns>
    [<Extension>]
    static member inline JsonAllOrdered(conn, tableName, orderFields) =
        Json.allOrdered tableName orderFields conn

    /// <summary>Retrieve a JSON document by its ID</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    /// <returns>The JSON document if found, an empty JSON document otherwise</returns>
    [<Extension>]
    static member inline JsonById<'TKey>(conn, tableName, docId: 'TKey) =
        Json.byId tableName docId conn

    /// <summary>Retrieve JSON documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>All JSON documents matching the given fields</returns>
    [<Extension>]
    static member inline JsonByFields(conn, tableName, howMatched, fields) =
        Json.byFields tableName howMatched fields conn

    /// <summary>
    /// Retrieve JSON documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields
    /// in the document
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All JSON documents matching the given fields, ordered by the other given fields</returns>
    [<Extension>]
    static member inline JsonByFieldsOrdered(conn, tableName, howMatched, queryFields, orderFields) =
        Json.byFieldsOrdered tableName howMatched queryFields orderFields conn

    /// <summary>Retrieve the first JSON document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>The first JSON document if found, an empty JSON document otherwise</returns>
    [<Extension>]
    static member inline JsonFirstByFields(conn, tableName, howMatched, fields) =
        Json.firstByFields tableName howMatched fields conn

    /// <summary>
    /// Retrieve the first JSON document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given
    /// fields in the document
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first JSON document (in order) if found, an empty JSON document otherwise</returns>
    [<Extension>]
    static member inline JsonFirstByFieldsOrdered(conn, tableName, howMatched, queryFields, orderFields) =
        Json.firstByFieldsOrdered tableName howMatched queryFields orderFields conn

    /// <summary>Write all JSON documents in the given table to the given <c>StreamWriter</c></summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    [<Extension>]
    static member inline WriteJsonAll(conn, tableName, writer) =
        Json.writeAll tableName writer conn

    /// <summary>
    /// Write all JSON all documents in the given table to the given <c>StreamWriter</c>, ordered by the given fields in
    /// the document
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<Extension>]
    static member inline WriteJsonAllOrdered(conn, tableName, writer, orderFields) =
        Json.writeAllOrdered tableName writer orderFields conn

    /// <summary>Write a JSON document to the given <c>StreamWriter</c> by its ID</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    [<Extension>]
    static member inline WriteJsonById<'TKey>(conn, tableName, writer, docId: 'TKey) =
        Json.writeById tableName writer docId conn

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    [<Extension>]
    static member inline WriteJsonByFields(conn, tableName, writer, howMatched, fields) =
        Json.writeByFields tableName writer howMatched fields conn

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)
    /// ordered by the given fields in the document
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<Extension>]
    static member inline WriteJsonByFieldsOrdered(conn, tableName, writer, howMatched, queryFields, orderFields) =
        Json.writeByFieldsOrdered tableName writer howMatched queryFields orderFields conn

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching JSON field comparisons
    /// (<c>-&gt;&gt; =</c>, etc.)
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    [<Extension>]
    static member inline WriteJsonFirstByFields(conn, tableName, writer, howMatched, fields) =
        Json.writeFirstByFields tableName writer howMatched fields conn

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching JSON field comparisons
    /// (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in the document
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<Extension>]
    static member inline WriteJsonFirstByFieldsOrdered(conn, tableName, writer, howMatched, queryFields, orderFields) =
        Json.writeFirstByFieldsOrdered tableName writer howMatched queryFields orderFields conn

    /// <summary>Update (replace) an entire document by its ID</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
    /// <param name="docId">The ID of the document to be updated (replaced)</param>
    /// <param name="document">The new document</param>
    [<Extension>]
    static member inline UpdateById<'TKey, 'TDoc>(conn, tableName, docId: 'TKey, document: 'TDoc) =
        Update.byId tableName docId document conn

    /// <summary>
    /// Update (replace) an entire document by its ID, using the provided function to obtain the ID from the document
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
    /// <param name="idFunc">The function to obtain the ID of the document</param>
    /// <param name="document">The new document</param>
    [<Extension>]
    static member inline UpdateByFunc<'TKey, 'TDoc>(
            conn, tableName, idFunc: System.Func<'TDoc, 'TKey>, document: 'TDoc) =
        Update.ByFunc(tableName, idFunc, document, conn)

    /// <summary>Patch a document by its ID</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which a document should be patched (may include schema)</param>
    /// <param name="docId">The ID of the document to patch</param>
    /// <param name="patch">The partial document to patch the existing document</param>
    [<Extension>]
    static member inline PatchById<'TKey, 'TPatch>(conn, tableName, docId: 'TKey, patch: 'TPatch) =
        Patch.byId tableName docId patch conn

    /// <summary>
    /// Patch documents using a JSON field comparison query in the <c>WHERE</c> clause (<c>-&gt;&gt; =</c>, etc.)
    /// </summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which documents should be patched (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <param name="patch">The partial document to patch the existing document</param>
    [<Extension>]
    static member inline PatchByFields<'TPatch>(conn, tableName, howMatched, fields, patch: 'TPatch) =
        Patch.byFields tableName howMatched fields patch conn

    /// <summary>Remove fields from a document by the document's ID</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which a document should be modified (may include schema)</param>
    /// <param name="docId">The ID of the document to modify</param>
    /// <param name="fieldNames">One or more field names to remove from the document</param>
    [<Extension>]
    static member inline RemoveFieldsById<'TKey>(conn, tableName, docId: 'TKey, fieldNames) =
        RemoveFields.byId tableName docId fieldNames conn

    /// <summary>Remove fields from documents via a comparison on JSON fields in the document</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which documents should be modified (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <param name="fieldNames">One or more field names to remove from the matching documents</param>
    [<Extension>]
    static member inline RemoveFieldsByFields(conn, tableName, howMatched, fields, fieldNames) =
        RemoveFields.byFields tableName howMatched fields fieldNames conn

    /// <summary>Delete a document by its ID</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which a document should be deleted (may include schema)</param>
    /// <param name="docId">The ID of the document to delete</param>
    [<Extension>]
    static member inline DeleteById<'TKey>(conn, tableName, docId: 'TKey) =
        Delete.byId tableName docId conn

    /// <summary>Delete documents by matching a JSON field comparison query (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="conn">The <c>SqliteConnection</c> on which to run the query</param>
    /// <param name="tableName">The table in which documents should be deleted (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    [<Extension>]
    static member inline DeleteByFields(conn, tableName, howMatched, fields) =
        Delete.byFields tableName howMatched fields conn
