namespace BitBadger.Documents.Sqlite

open Microsoft.Data.Sqlite

/// <summary>Commands to execute custom SQL queries</summary>
[<RequireQualifiedAccess>]
module Custom =

    /// <summary>Execute a query that returns a list of results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns>A list of results for the given query</returns>
    [<CompiledName "FSharpList">]
    let list<'TDoc> query parameters (mapFunc: SqliteDataReader -> 'TDoc) =
        use conn = Configuration.dbConn ()
        WithConn.Custom.list<'TDoc> query parameters mapFunc conn

    /// <summary>Execute a query that returns a list of results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns>A list of results for the given query</returns>
    let List<'TDoc>(query, parameters, mapFunc: System.Func<SqliteDataReader, 'TDoc>) =
        use conn = Configuration.dbConn ()
        WithConn.Custom.List<'TDoc>(query, parameters, mapFunc, conn)

    /// <summary>Execute a query that returns a JSON array of results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>A JSON array of results for the given query</returns>
    [<CompiledName "FSharpJsonArray">]
    let jsonArray query parameters mapFunc =
        use conn = Configuration.dbConn ()
        WithConn.Custom.jsonArray query parameters mapFunc conn

    /// <summary>Execute a query that returns a JSON array of results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>A JSON array of results for the given query</returns>
    let JsonArray(query, parameters, mapFunc) =
        use conn = Configuration.dbConn ()
        WithConn.Custom.JsonArray(query, parameters, mapFunc, conn)

    /// <summary>Execute a query, writing its results to the given <c>StreamWriter</c></summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    [<CompiledName "FSharpWriteJsonArray">]
    let writeJsonArray query parameters writer mapFunc =
        use conn = Configuration.dbConn ()
        WithConn.Custom.writeJsonArray query parameters writer mapFunc conn

    /// <summary>Execute a query, writing its results to the given <c>StreamWriter</c></summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    let WriteJsonArray(query, parameters, writer, mapFunc) =
        use conn = Configuration.dbConn ()
        WithConn.Custom.WriteJsonArray(query, parameters, writer, mapFunc, conn)
        
    /// <summary>Execute a query that returns one or no results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns><c>Some</c> with the first matching result, or <c>None</c> if not found</returns>
    [<CompiledName "FSharpSingle">]
    let single<'TDoc> query parameters (mapFunc: SqliteDataReader -> 'TDoc) =
        use conn = Configuration.dbConn ()
        WithConn.Custom.single<'TDoc> query parameters mapFunc conn

    /// <summary>Execute a query that returns one or no results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function between the document and the domain item</param>
    /// <returns>The first matching result, or <c>null</c> if not found</returns>
    let Single<'TDoc when 'TDoc: null and 'TDoc: not struct>(
            query, parameters, mapFunc: System.Func<SqliteDataReader, 'TDoc>) =
        use conn = Configuration.dbConn ()
        WithConn.Custom.Single<'TDoc>(query, parameters, mapFunc, conn)

    /// <summary>Execute a query that returns one or no JSON documents</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>The JSON document with the first matching result, or an empty document if not found</returns>
    [<CompiledName "FSharpJsonSingle">]
    let jsonSingle query parameters mapFunc =
        use conn = Configuration.dbConn ()
        WithConn.Custom.jsonSingle query parameters mapFunc conn

    /// <summary>Execute a query that returns one or no JSON documents</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to extract the document</param>
    /// <returns>The JSON document with the first matching result, or an empty document if not found</returns>
    let JsonSingle(query, parameters, mapFunc) =
        use conn = Configuration.dbConn ()
        WithConn.Custom.JsonSingle(query, parameters, mapFunc, conn)

    /// <summary>Execute a query that returns no results</summary>
    /// <param name="query">The query to retrieve the results</param>
    /// <param name="parameters">Parameters to use for the query</param>
    [<CompiledName "NonQuery">]
    let nonQuery query parameters =
        use conn = Configuration.dbConn ()
        WithConn.Custom.nonQuery query parameters conn

    /// <summary>Execute a query that returns a scalar value</summary>
    /// <param name="query">The query to retrieve the value</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to obtain the value</param>
    /// <returns>The scalar value for the query</returns>
    [<CompiledName "FSharpScalar">]
    let scalar<'T when 'T: struct> query parameters (mapFunc: SqliteDataReader -> 'T) =
        use conn = Configuration.dbConn ()
        WithConn.Custom.scalar<'T> query parameters mapFunc conn

    /// <summary>Execute a query that returns a scalar value</summary>
    /// <param name="query">The query to retrieve the value</param>
    /// <param name="parameters">Parameters to use for the query</param>
    /// <param name="mapFunc">The mapping function to obtain the value</param>
    /// <returns>The scalar value for the query</returns>
    let Scalar<'T when 'T: struct>(query, parameters, mapFunc: System.Func<SqliteDataReader, 'T>) =
        use conn = Configuration.dbConn ()
        WithConn.Custom.Scalar<'T>(query, parameters, mapFunc, conn)


/// <summary>Functions to create tables and indexes</summary>
[<RequireQualifiedAccess>]
module Definition =

    /// <summary>Create a document table</summary>
    /// <param name="name">The table whose existence should be ensured (may include schema)</param>
    [<CompiledName "EnsureTable">]
    let ensureTable name =
        use conn = Configuration.dbConn ()
        WithConn.Definition.ensureTable name conn

    /// <summary>Create an index on field(s) within documents in the specified table</summary>
    /// <param name="tableName">The table to be indexed (may include schema)</param>
    /// <param name="indexName">The name of the index to create</param>
    /// <param name="fields">One or more fields to be indexed</param>
    [<CompiledName "EnsureFieldIndex">]
    let ensureFieldIndex tableName indexName fields =
        use conn = Configuration.dbConn ()
        WithConn.Definition.ensureFieldIndex tableName indexName fields conn


/// <summary>Document insert/save functions</summary>
[<AutoOpen>]
module Document =

    /// <summary>Insert a new document</summary>
    /// <param name="tableName">The table into which the document should be inserted (may include schema)</param>
    /// <param name="document">The document to be inserted</param>
    [<CompiledName "Insert">]
    let insert<'TDoc> tableName (document: 'TDoc) =
        use conn = Configuration.dbConn ()
        WithConn.Document.insert tableName document conn

    /// <summary>Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")</summary>
    /// <param name="tableName">The table into which the document should be saved (may include schema)</param>
    /// <param name="document">The document to be saved</param>
    [<CompiledName "Save">]
    let save<'TDoc> tableName (document: 'TDoc) =
        use conn = Configuration.dbConn ()
        WithConn.Document.save tableName document conn


/// <summary>Commands to count documents</summary>
[<RequireQualifiedAccess>]
module Count =

    /// <summary>Count all documents in a table</summary>
    /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
    /// <returns>The count of the documents in the table</returns>
    [<CompiledName "All">]
    let all tableName =
        use conn = Configuration.dbConn ()
        WithConn.Count.all tableName conn

    /// <summary>Count matching documents using JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table in which documents should be counted (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>The count of matching documents in the table</returns>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        use conn = Configuration.dbConn ()
        WithConn.Count.byFields tableName howMatched fields conn


/// <summary>Commands to determine if documents exist</summary>
[<RequireQualifiedAccess>]
module Exists =

    /// <summary>Determine if a document exists for the given ID</summary>
    /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
    /// <param name="docId">The ID of the document whose existence should be checked</param>
    /// <returns>True if a document exists, false if not</returns>
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) =
        use conn = Configuration.dbConn ()
        WithConn.Exists.byId tableName docId conn

    /// <summary>Determine if a document exists using JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table in which existence should be checked (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>True if any matching documents exist, false if not</returns>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        use conn = Configuration.dbConn ()
        WithConn.Exists.byFields tableName howMatched fields conn


/// <summary>Commands to retrieve documents</summary>
[<RequireQualifiedAccess>]
module Find =

    /// <summary>Retrieve all documents in the given table</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <returns>All documents from the given table</returns>
    [<CompiledName "FSharpAll">]
    let all<'TDoc> tableName =
        use conn = Configuration.dbConn ()
        WithConn.Find.all<'TDoc> tableName conn

    /// <summary>Retrieve all documents in the given table</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <returns>All documents from the given table</returns>
    let All<'TDoc> tableName =
        use conn = Configuration.dbConn ()
        WithConn.Find.All<'TDoc>(tableName, conn)

    /// <summary>Retrieve all documents in the given table ordered by the given fields in the document</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents from the given table, ordered by the given fields</returns>
    [<CompiledName "FSharpAllOrdered">]
    let allOrdered<'TDoc> tableName orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Find.allOrdered<'TDoc> tableName orderFields conn

    /// <summary>Retrieve all documents in the given table ordered by the given fields in the document</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents from the given table, ordered by the given fields</returns>
    let AllOrdered<'TDoc> tableName orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Find.AllOrdered<'TDoc>(tableName, orderFields, conn)

    /// <summary>Retrieve a document by its ID</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    /// <returns><c>Some</c> with the document if found, <c>None</c> otherwise</returns>
    [<CompiledName "FSharpById">]
    let byId<'TKey, 'TDoc> tableName docId =
        use conn = Configuration.dbConn ()
        WithConn.Find.byId<'TKey, 'TDoc> tableName docId conn

    /// <summary>Retrieve a document by its ID</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    /// <returns>The document if found, <c>null</c> otherwise</returns>
    let ById<'TKey, 'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, docId) =
        use conn = Configuration.dbConn ()
        WithConn.Find.ById<'TKey, 'TDoc>(tableName, docId, conn)

    /// <summary>Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>All documents matching the given fields</returns>
    [<CompiledName "FSharpByFields">]
    let byFields<'TDoc> tableName howMatched fields =
        use conn = Configuration.dbConn ()
        WithConn.Find.byFields<'TDoc> tableName howMatched fields conn

    /// <summary>Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>All documents matching the given fields</returns>
    let ByFields<'TDoc>(tableName, howMatched, fields) =
        use conn = Configuration.dbConn ()
        WithConn.Find.ByFields<'TDoc>(tableName, howMatched, fields, conn)

    /// <summary>
    /// Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given fields, ordered by the other given fields</returns>
    [<CompiledName "FSharpByFieldsOrdered">]
    let byFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Find.byFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields conn

    /// <summary>
    /// Retrieve documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in the
    /// document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All documents matching the given fields, ordered by the other given fields</returns>
    let ByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields) =
        use conn = Configuration.dbConn ()
        WithConn.Find.ByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields, conn)

    /// <summary>Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns><c>Some</c> with the first document, or <c>None</c> if not found</returns>
    [<CompiledName "FSharpFirstByFields">]
    let firstByFields<'TDoc> tableName howMatched fields =
        use conn = Configuration.dbConn ()
        WithConn.Find.firstByFields<'TDoc> tableName howMatched fields conn

    /// <summary>Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>The first document, or <c>null</c> if not found</returns>
    let FirstByFields<'TDoc when 'TDoc: null and 'TDoc: not struct>(tableName, howMatched, fields) =
        use conn = Configuration.dbConn ()
        WithConn.Find.FirstByFields<'TDoc>(tableName, howMatched, fields, conn)

    /// <summary>
    /// Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given
    /// fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>
    /// <c>Some</c> with the first document ordered by the given fields, or <c>None</c> if not found
    /// </returns>
    [<CompiledName "FSharpFirstByFieldsOrdered">]
    let firstByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Find.firstByFieldsOrdered<'TDoc> tableName howMatched queryFields orderFields conn

    /// <summary>
    /// Retrieve the first document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given
    /// fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first document ordered by the given fields, or <c>null</c> if not found</returns>
    let FirstByFieldsOrdered<'TDoc when 'TDoc: null and 'TDoc: not struct>(
            tableName, howMatched, queryFields, orderFields) =
        use conn = Configuration.dbConn ()
        WithConn.Find.FirstByFieldsOrdered<'TDoc>(tableName, howMatched, queryFields, orderFields, conn)


/// <summary>Commands to retrieve documents as raw JSON</summary>
[<RequireQualifiedAccess>]
module Json =

    /// <summary>Retrieve all JSON documents in the given table</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <returns>All JSON documents from the given table</returns>
    [<CompiledName "All">]
    let all tableName =
        use conn = Configuration.dbConn ()
        WithConn.Json.all tableName conn

    /// <summary>Retrieve all JSON documents in the given table ordered by the given fields in the document</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All JSON documents from the given table, ordered by the given fields</returns>
    [<CompiledName "AllOrdered">]
    let allOrdered tableName orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Json.allOrdered tableName orderFields conn

    /// <summary>Retrieve a JSON document by its ID</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    /// <returns>The JSON document if found, an empty JSON document otherwise</returns>
    [<CompiledName "ById">]
    let byId<'TKey> tableName (docId: 'TKey) =
        use conn = Configuration.dbConn ()
        WithConn.Json.byId tableName docId conn

    /// <summary>Retrieve JSON documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>All JSON documents matching the given fields</returns>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        use conn = Configuration.dbConn ()
        WithConn.Json.byFields tableName howMatched fields conn

    /// <summary>
    /// Retrieve JSON documents matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields
    /// in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>All JSON documents matching the given fields, ordered by the other given fields</returns>
    [<CompiledName "ByFieldsOrdered">]
    let byFieldsOrdered tableName howMatched queryFields orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Json.byFieldsOrdered tableName howMatched queryFields orderFields conn

    /// <summary>Retrieve the first JSON document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <returns>The first JSON document if found, an empty JSON document otherwise</returns>
    [<CompiledName "FirstByFields">]
    let firstByFields tableName howMatched fields =
        use conn = Configuration.dbConn ()
        WithConn.Json.firstByFields tableName howMatched fields conn

    /// <summary>
    /// Retrieve the first JSON document matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.) ordered by the given
    /// fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    /// <returns>The first JSON document (in order) if found, an empty JSON document otherwise</returns>
    [<CompiledName "FirstByFieldsOrdered">]
    let firstByFieldsOrdered tableName howMatched queryFields orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Json.firstByFieldsOrdered tableName howMatched queryFields orderFields conn

    /// <summary>Write all JSON documents in the given table to the given <c>StreamWriter</c></summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    [<CompiledName "WriteAll">]
    let writeAll tableName writer =
        use conn = Configuration.dbConn ()
        WithConn.Json.writeAll tableName writer conn

    /// <summary>
    /// Write all JSON all documents in the given table to the given <c>StreamWriter</c>, ordered by the given fields in
    /// the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteAllOrdered">]
    let writeAllOrdered tableName writer orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Json.writeAllOrdered tableName writer orderFields conn

    /// <summary>Write a JSON document to the given <c>StreamWriter</c> by its ID</summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="docId">The ID of the document to retrieve</param>
    [<CompiledName "WriteById">]
    let writeById<'TKey> tableName writer (docId: 'TKey) =
        use conn = Configuration.dbConn ()
        WithConn.Json.writeById tableName writer docId conn

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    [<CompiledName "WriteByFields">]
    let writeByFields tableName writer howMatched fields =
        use conn = Configuration.dbConn ()
        WithConn.Json.writeByFields tableName writer howMatched fields conn

    /// <summary>
    /// Write JSON documents to the given <c>StreamWriter</c> matching JSON field comparisons (<c>-&gt;&gt; =</c>, etc.)
    /// ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which documents should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteByFieldsOrdered">]
    let writeByFieldsOrdered tableName writer howMatched queryFields orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Json.writeByFieldsOrdered tableName writer howMatched queryFields orderFields conn

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching JSON field comparisons
    /// (<c>-&gt;&gt; =</c>, etc.)
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    [<CompiledName "WriteFirstByFields">]
    let writeFirstByFields tableName writer howMatched fields =
        use conn = Configuration.dbConn ()
        WithConn.Json.writeFirstByFields tableName writer howMatched fields conn

    /// <summary>
    /// Write the first JSON document to the given <c>StreamWriter</c> matching JSON field comparisons
    /// (<c>-&gt;&gt; =</c>, etc.) ordered by the given fields in the document
    /// </summary>
    /// <param name="tableName">The table from which a document should be retrieved (may include schema)</param>
    /// <param name="writer">The StreamWriter to which the results should be written</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="queryFields">The field conditions to match</param>
    /// <param name="orderFields">Fields by which the results should be ordered</param>
    [<CompiledName "WriteFirstByFieldsOrdered">]
    let writeFirstByFieldsOrdered tableName writer howMatched queryFields orderFields =
        use conn = Configuration.dbConn ()
        WithConn.Json.writeFirstByFieldsOrdered tableName writer howMatched queryFields orderFields conn


/// <summary>Commands to update documents</summary>
[<RequireQualifiedAccess>]
module Update =

    /// <summary>Update (replace) an entire document by its ID</summary>
    /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
    /// <param name="docId">The ID of the document to be updated (replaced)</param>
    /// <param name="document">The new document</param>
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) (document: 'TDoc) =
        use conn = Configuration.dbConn ()
        WithConn.Update.byId tableName docId document conn

    /// <summary>
    /// Update (replace) an entire document by its ID, using the provided function to obtain the ID from the document
    /// </summary>
    /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
    /// <param name="idFunc">The function to obtain the ID of the document</param>
    /// <param name="document">The new document</param>
    [<CompiledName "FSharpByFunc">]
    let byFunc tableName (idFunc: 'TDoc -> 'TKey) (document: 'TDoc) =
        use conn = Configuration.dbConn ()
        WithConn.Update.byFunc tableName idFunc document conn

    /// <summary>
    /// Update (replace) an entire document by its ID, using the provided function to obtain the ID from the document
    /// </summary>
    /// <param name="tableName">The table in which a document should be updated (may include schema)</param>
    /// <param name="idFunc">The function to obtain the ID of the document</param>
    /// <param name="document">The new document</param>
    let ByFunc(tableName, idFunc: System.Func<'TDoc, 'TKey>, document: 'TDoc) =
        use conn = Configuration.dbConn ()
        WithConn.Update.ByFunc(tableName, idFunc, document, conn)


/// <summary>Commands to patch (partially update) documents</summary>
[<RequireQualifiedAccess>]
module Patch =

    /// <summary>Patch a document by its ID</summary>
    /// <param name="tableName">The table in which a document should be patched (may include schema)</param>
    /// <param name="docId">The ID of the document to patch</param>
    /// <param name="patch">The partial document to patch the existing document</param>
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) (patch: 'TPatch) =
        use conn = Configuration.dbConn ()
        WithConn.Patch.byId tableName docId patch conn

    /// <summary>
    /// Patch documents using a JSON field comparison query in the <c>WHERE</c> clause (<c>-&gt;&gt; =</c>, etc.)
    /// </summary>
    /// <param name="tableName">The table in which documents should be patched (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <param name="patch">The partial document to patch the existing document</param>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields (patch: 'TPatch) =
        use conn = Configuration.dbConn ()
        WithConn.Patch.byFields tableName howMatched fields patch conn


/// <summary>Commands to remove fields from documents</summary>
[<RequireQualifiedAccess>]
module RemoveFields =

    /// <summary>Remove fields from a document by the document's ID</summary>
    /// <param name="tableName">The table in which a document should be modified (may include schema)</param>
    /// <param name="docId">The ID of the document to modify</param>
    /// <param name="fieldNames">One or more field names to remove from the document</param>
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) fieldNames =
        use conn = Configuration.dbConn ()
        WithConn.RemoveFields.byId tableName docId fieldNames conn

    /// <summary>Remove fields from documents via a comparison on JSON fields in the document</summary>
    /// <param name="tableName">The table in which documents should be modified (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    /// <param name="fieldNames">One or more field names to remove from the matching documents</param>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields fieldNames =
        use conn = Configuration.dbConn ()
        WithConn.RemoveFields.byFields tableName howMatched fields fieldNames conn


/// <summary>Commands to delete documents</summary>
[<RequireQualifiedAccess>]
module Delete =

    /// <summary>Delete a document by its ID</summary>
    /// <param name="tableName">The table in which a document should be deleted (may include schema)</param>
    /// <param name="docId">The ID of the document to delete</param>
    [<CompiledName "ById">]
    let byId tableName (docId: 'TKey) =
        use conn = Configuration.dbConn ()
        WithConn.Delete.byId tableName docId conn

    /// <summary>Delete documents by matching a JSON field comparison query (<c>-&gt;&gt; =</c>, etc.)</summary>
    /// <param name="tableName">The table in which documents should be deleted (may include schema)</param>
    /// <param name="howMatched">Whether to match any or all of the field conditions</param>
    /// <param name="fields">The field conditions to match</param>
    [<CompiledName "ByFields">]
    let byFields tableName howMatched fields =
        use conn = Configuration.dbConn ()
        WithConn.Delete.byFields tableName howMatched fields conn
