module SqliteExtensionTests

open System.IO
open System.Text.Json
open BitBadger.Documents
open BitBadger.Documents.Sqlite
open BitBadger.Documents.Tests
open Expecto
open Microsoft.Data.Sqlite
open Types

/// Integration tests for the F# extensions on the SqliteConnection data type
let integrationTests =
    let loadDocs (conn: SqliteConnection) = backgroundTask {
        for doc in testDocuments do do! conn.insert SqliteDb.TableName doc
    }

    /// Set up a stream writer for a test
    let writeStream (stream: Stream) =
        let writer = new StreamWriter(stream)
        writer.AutoFlush <- true
        writer

    /// Get the text of the given stream
    let streamText (stream: Stream) =
        stream.Position <- 0L
        use reader = new StreamReader(stream)
        reader.ReadToEnd()

    /// Verify a JSON array begins with "[" and ends with "]"
    let verifyBeginEnd json =
        Expect.stringStarts json "[" "The array should have started with `[`"
        Expect.stringEnds   json "]" "The array should have ended with `]`"

    /// Verify an empty JSON array
    let verifyEmpty json =
        Expect.equal json "[]" "There should be no documents returned"

    /// Verify an empty JSON document
    let verifyNoDoc json =
        Expect.equal json "{}" "There should be no document returned"

    /// Verify the presence of any of the given documents in the given JSON
    let verifyAny (json: string) (docs: string list) =
        match docs |> List.tryFind json.Contains with
        | Some _ -> ()
        | None ->
            let theDocs = docs |> String.concat " | "
            Expect.isTrue false $"Could not find any of |{theDocs}| in {json}"

    testList "Sqlite.Extensions" [
        testTask "ensureTable succeeds" {
            use! db   = SqliteDb.BuildDb()
            use  conn = Configuration.dbConn ()
            let itExists (name: string) =
                conn.customScalar
                    $"SELECT EXISTS (SELECT 1 FROM {SqliteDb.Catalog} WHERE name = @name) AS it"
                    [ SqliteParameter("@name", name) ]
                    toExists

            let! exists     = itExists "ensured"
            let! alsoExists = itExists "idx_ensured_key"
            Expect.isFalse exists     "The table should not exist already"
            Expect.isFalse alsoExists "The key index should not exist already"

            do! conn.ensureTable "ensured"
            let! exists'     = itExists "ensured"
            let! alsoExists' = itExists "idx_ensured_key"
            Expect.isTrue exists'    "The table should now exist"
            Expect.isTrue alsoExists' "The key index should now exist"
        }
        testTask "ensureFieldIndex succeeds" {
            use! db   = SqliteDb.BuildDb()
            use  conn = Configuration.dbConn ()
            let indexExists () =
                conn.customScalar
                    $"SELECT EXISTS (SELECT 1 FROM {SqliteDb.Catalog} WHERE name = 'idx_ensured_test') AS it"
                    []
                    toExists

            let! exists = indexExists ()
            Expect.isFalse exists "The index should not exist already"

            do! conn.ensureTable      "ensured"
            do! conn.ensureFieldIndex "ensured" "test" [ "Name"; "Age" ]
            let! exists' = indexExists ()
            Expect.isTrue exists' "The index should now exist"
        }
        testList "insert" [
            testTask "succeeds" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                let! before = conn.findAll<SubDocument> SqliteDb.TableName
                Expect.equal before [] "There should be no documents in the table"

                let testDoc = { emptyDoc with Id = "turkey"; Sub = Some { Foo = "gobble"; Bar = "gobble" } }
                do! conn.insert SqliteDb.TableName testDoc
                let! after = conn.findAll<JsonDocument> SqliteDb.TableName
                Expect.equal after [ testDoc ] "There should have been one document inserted"
            }
            testTask "fails for duplicate key" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! conn.insert SqliteDb.TableName { emptyDoc with Id = "test" }
                Expect.throws
                    (fun () ->
                        conn.insert SqliteDb.TableName {emptyDoc with Id = "test" }
                        |> Async.AwaitTask
                        |> Async.RunSynchronously)
                    "An exception should have been raised for duplicate document ID insert"
            }
        ]
        testList "save" [
            testTask "succeeds when a document is inserted" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                let! before = conn.findAll<JsonDocument> SqliteDb.TableName
                Expect.equal before [] "There should be no documents in the table"

                let testDoc = { emptyDoc with Id = "test"; Sub = Some { Foo = "a"; Bar = "b" } }
                do! conn.save SqliteDb.TableName testDoc
                let! after = conn.findAll<JsonDocument> SqliteDb.TableName
                Expect.equal after [ testDoc ] "There should have been one document inserted"
            }
            testTask "succeeds when a document is updated" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                let testDoc = { emptyDoc with Id = "test"; Sub = Some { Foo = "a"; Bar = "b" } }
                do! conn.insert SqliteDb.TableName testDoc

                let! before = conn.findById<string, JsonDocument> SqliteDb.TableName "test"
                if Option.isNone before then Expect.isTrue false "There should have been a document returned"
                Expect.equal before.Value testDoc "The document is not correct"

                let upd8Doc = { testDoc with Sub = Some { Foo = "c"; Bar = "d" } }
                do! conn.save SqliteDb.TableName upd8Doc
                let! after = conn.findById<string, JsonDocument> SqliteDb.TableName "test"
                if Option.isNone after then
                    Expect.isTrue false "There should have been a document returned post-update"
                Expect.equal after.Value upd8Doc "The updated document is not correct"
            }
        ]
        testTask "countAll succeeds" {
            use! db   = SqliteDb.BuildDb()
            use  conn = Configuration.dbConn ()
            do! loadDocs conn

            let! theCount = conn.countAll SqliteDb.TableName
            Expect.equal theCount 5L "There should have been 5 matching documents"
        }
        testTask "countByFields succeeds" {
            use! db   = SqliteDb.BuildDb()
            use  conn = Configuration.dbConn ()
            do! loadDocs conn

            let! theCount = conn.countByFields SqliteDb.TableName Any [ Field.Equal "Value" "purple" ]
            Expect.equal theCount 2L "There should have been 2 matching documents"
        }
        testList "existsById" [
            testTask "succeeds when a document exists" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! exists = conn.existsById SqliteDb.TableName "three"
                Expect.isTrue exists "There should have been an existing document"
            }
            testTask "succeeds when a document does not exist" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! exists = conn.existsById SqliteDb.TableName "seven"
                Expect.isFalse exists "There should not have been an existing document"
            }
        ]
        testList "existsByFields" [
            testTask "succeeds when documents exist" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! exists = conn.existsByFields SqliteDb.TableName Any [ Field.Equal "NumValue" 10 ]
                Expect.isTrue exists "There should have been existing documents"
            }
            testTask "succeeds when no matching documents exist" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! exists = conn.existsByFields SqliteDb.TableName Any [ Field.Equal "Nothing" "none" ]
                Expect.isFalse exists "There should not have been any existing documents"
            }
        ]
        testList "findAll" [
            testTask "succeeds when there is data" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()

                do! insert SqliteDb.TableName { Foo = "one"; Bar = "two" }
                do! insert SqliteDb.TableName { Foo = "three"; Bar = "four" }
                do! insert SqliteDb.TableName { Foo = "five"; Bar = "six" }

                let! results = conn.findAll<SubDocument> SqliteDb.TableName
                let expected = [
                    { Foo = "one"; Bar = "two" }
                    { Foo = "three"; Bar = "four" }
                    { Foo = "five"; Bar = "six" }
                ]
                Expect.equal results expected "There should have been 3 documents returned"
            }
            testTask "succeeds when there is no data" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                let! results = conn.findAll<SubDocument> SqliteDb.TableName
                Expect.equal results [] "There should have been no documents returned"
            }
        ]
        testList "findAllOrdered" [
            testTask "succeeds when ordering numerically" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! results = conn.findAllOrdered<JsonDocument> SqliteDb.TableName [ Field.Named "n:NumValue" ]
                Expect.hasLength results 5 "There should have been 5 documents returned"
                Expect.equal
                    (results |> List.map _.Id |> String.concat "|")
                    "one|three|two|four|five"
                    "The documents were not ordered correctly"
            }
            testTask "succeeds when ordering numerically descending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! results = conn.findAllOrdered<JsonDocument> SqliteDb.TableName [ Field.Named "n:NumValue DESC" ]
                Expect.hasLength results 5 "There should have been 5 documents returned"
                Expect.equal
                    (results |> List.map _.Id |> String.concat "|")
                    "five|four|two|three|one"
                    "The documents were not ordered correctly"
            }
            testTask "succeeds when ordering alphabetically" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! results = conn.findAllOrdered<JsonDocument> SqliteDb.TableName [ Field.Named "Id DESC" ]
                Expect.hasLength results 5 "There should have been 5 documents returned"
                Expect.equal
                    (results |> List.map _.Id |> String.concat "|")
                    "two|three|one|four|five"
                    "The documents were not ordered correctly"
            }
        ]
        testList "findById" [
            testTask "succeeds when a document is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! doc = conn.findById<string, JsonDocument> SqliteDb.TableName "two"
                Expect.isSome doc "There should have been a document returned"
                Expect.equal doc.Value.Id "two" "The incorrect document was returned"
            }
            testTask "succeeds when a document is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! doc = conn.findById<string, JsonDocument> SqliteDb.TableName "three hundred eighty-seven"
                Expect.isNone doc "There should not have been a document returned"
            }
        ]
        testList "findByFields" [
            testTask "succeeds when documents are found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! docs = conn.findByFields<JsonDocument> SqliteDb.TableName Any [ Field.Equal "Sub.Foo" "green" ]
                Expect.hasLength docs 2 "There should have been two documents returned"
            }
            testTask "succeeds when documents are not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! docs = conn.findByFields<JsonDocument> SqliteDb.TableName Any [ Field.Equal "Value" "mauve" ]
                Expect.isEmpty docs "There should have been no documents returned"
            }
        ]
        testList "findByFieldsOrdered" [
            testTask "succeeds when sorting ascending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! docs =
                    conn.findByFieldsOrdered<JsonDocument>
                        SqliteDb.TableName Any [ Field.Greater "NumValue" 15 ] [ Field.Named "Id" ]
                Expect.equal
                    (docs |> List.map _.Id |> String.concat "|") "five|four" "The documents were not ordered correctly"
            }
            testTask "succeeds when sorting descending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! docs =
                    conn.findByFieldsOrdered<JsonDocument>
                        SqliteDb.TableName Any [ Field.Greater "NumValue" 15 ] [ Field.Named "Id DESC" ]
                Expect.equal
                    (docs |> List.map _.Id |> String.concat "|") "four|five" "The documents were not ordered correctly"
            }
        ]
        testList "findFirstByFields" [
            testTask "succeeds when a document is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! doc = conn.findFirstByFields<JsonDocument> SqliteDb.TableName Any [ Field.Equal "Value" "another" ]
                Expect.isSome doc "There should have been a document returned"
                Expect.equal doc.Value.Id "two" "The incorrect document was returned"
            }
            testTask "succeeds when multiple documents are found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! doc = conn.findFirstByFields<JsonDocument> SqliteDb.TableName Any [ Field.Equal "Sub.Foo" "green" ]
                Expect.isSome doc "There should have been a document returned"
                Expect.contains [ "two"; "four" ] doc.Value.Id "An incorrect document was returned"
            }
            testTask "succeeds when a document is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! doc = conn.findFirstByFields<JsonDocument> SqliteDb.TableName Any [ Field.Equal "Value" "absent" ]
                Expect.isNone doc "There should not have been a document returned"
            }
        ]
        testList "findFirstByFieldsOrdered" [
            testTask "succeeds when sorting ascending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! doc =
                    conn.findFirstByFieldsOrdered<JsonDocument>
                        SqliteDb.TableName Any [ Field.Equal "Sub.Foo" "green" ] [ Field.Named "Sub.Bar" ]
                Expect.isSome doc "There should have been a document returned"
                Expect.equal "two" doc.Value.Id "An incorrect document was returned"
            }
            testTask "succeeds when sorting descending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! doc =
                    conn.findFirstByFieldsOrdered<JsonDocument>
                        SqliteDb.TableName Any [ Field.Equal "Sub.Foo" "green" ] [ Field.Named "Sub.Bar DESC" ]
                Expect.isSome doc "There should have been a document returned"
                Expect.equal "four" doc.Value.Id "An incorrect document was returned"
            }
        ]
        testList "jsonAll" [
            testTask "succeeds when there is data" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()

                do! conn.insert SqliteDb.TableName { Foo = "one"; Bar = "two" }
                do! conn.insert SqliteDb.TableName { Foo = "three"; Bar = "four" }
                do! conn.insert SqliteDb.TableName { Foo = "five"; Bar = "six" }

                let! json = conn.jsonAll SqliteDb.TableName
                verifyBeginEnd json
                Expect.stringContains json """{"Foo":"one","Bar":"two"}"""    "The first document was not found"
                Expect.stringContains json """{"Foo":"three","Bar":"four"}""" "The second document was not found"
                Expect.stringContains json """{"Foo":"five","Bar":"six"}"""   "The third document was not found"
            }
            testTask "succeeds when there is no data" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                let! json = conn.jsonAll SqliteDb.TableName
                verifyEmpty json
            }
        ]
        testList "jsonAllOrdered" [
            testTask "succeeds when ordering numerically" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonAllOrdered SqliteDb.TableName [ Field.Named "n:NumValue" ]
                Expect.equal
                    json
                    $"[{JsonDocument.one},{JsonDocument.three},{JsonDocument.two},{JsonDocument.four},{JsonDocument.five}]"
                    "The documents were not ordered correctly"
            }
            testTask "succeeds when ordering numerically descending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonAllOrdered SqliteDb.TableName [ Field.Named "n:NumValue DESC" ]
                Expect.equal
                    json
                    $"[{JsonDocument.five},{JsonDocument.four},{JsonDocument.two},{JsonDocument.three},{JsonDocument.one}]"
                    "The documents were not ordered correctly"
            }
            testTask "succeeds when ordering alphabetically" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonAllOrdered SqliteDb.TableName [ Field.Named "Id DESC" ]
                Expect.equal
                    json
                    $"[{JsonDocument.two},{JsonDocument.three},{JsonDocument.one},{JsonDocument.four},{JsonDocument.five}]"
                    "The documents were not ordered correctly"
            }
        ]
        testList "jsonById" [
            testTask "succeeds when a document is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonById SqliteDb.TableName "two"
                Expect.equal json JsonDocument.two "The incorrect document was returned"
            }
            testTask "succeeds when a document is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonById SqliteDb.TableName "three hundred eighty-seven"
                verifyNoDoc json
            }
        ]
        testList "jsonByFields" [
            testTask "succeeds when documents are found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonByFields SqliteDb.TableName Any [ Field.Greater "NumValue" 15 ]
                verifyBeginEnd json
                Expect.stringContains json JsonDocument.four "Document `four` should have been returned"
                Expect.stringContains json JsonDocument.five "Document `five` should have been returned"
            }
            testTask "succeeds when documents are found using IN with numeric field" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonByFields SqliteDb.TableName All [ Field.In "NumValue" [ 2; 4; 6; 8 ] ]
                Expect.equal json $"[{JsonDocument.three}]" "There should have been one document returned"
            }
            testTask "succeeds when documents are not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonByFields SqliteDb.TableName Any [ Field.Greater "NumValue" 100 ]
                verifyEmpty json
            }
            testTask "succeeds for InArray when matching documents exist" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! conn.ensureTable SqliteDb.TableName
                for doc in ArrayDocument.TestDocuments do do! conn.insert SqliteDb.TableName doc

                let! json =
                    conn.jsonByFields SqliteDb.TableName All [ Field.InArray "Values" SqliteDb.TableName [ "c" ] ]
                verifyBeginEnd json
                Expect.stringContains
                    json """{"Id":"first","Values":["a","b","c"]}""" "Document `first` should have been returned"
                Expect.stringContains
                    json """{"Id":"second","Values":["c","d","e"]}""" "Document `second` should have been returned"
            }
            testTask "succeeds for InArray when no matching documents exist" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! conn.ensureTable SqliteDb.TableName
                for doc in ArrayDocument.TestDocuments do do! conn.insert SqliteDb.TableName doc

                let! json =
                    conn.jsonByFields SqliteDb.TableName All [ Field.InArray "Values" SqliteDb.TableName [ "j" ] ]
                verifyEmpty json
            }
        ]
        testList "jsonByFieldsOrdered" [
            testTask "succeeds when sorting ascending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json =
                    conn.jsonByFieldsOrdered SqliteDb.TableName Any [ Field.Greater "NumValue" 15 ] [ Field.Named "Id" ]
                Expect.equal json $"[{JsonDocument.five},{JsonDocument.four}]" "Incorrect documents were returned"
            }
            testTask "succeeds when sorting descending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json =
                    conn.jsonByFieldsOrdered
                        SqliteDb.TableName Any [ Field.Greater "NumValue" 15 ] [ Field.Named "Id DESC" ]
                Expect.equal json $"[{JsonDocument.four},{JsonDocument.five}]" "Incorrect documents were returned"
            }
            testTask "succeeds when sorting case-sensitively" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json =
                    conn.jsonByFieldsOrdered
                        SqliteDb.TableName All [ Field.LessOrEqual "NumValue" 10 ] [ Field.Named "Value" ]
                Expect.equal
                    json
                    $"[{JsonDocument.three},{JsonDocument.one},{JsonDocument.two}]"
                    "Documents not ordered correctly"
            }
            testTask "succeeds when sorting case-insensitively" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json =
                    conn.jsonByFieldsOrdered
                        SqliteDb.TableName All [ Field.LessOrEqual "NumValue" 10 ] [ Field.Named "i:Value" ]
                Expect.equal
                    json
                    $"[{JsonDocument.three},{JsonDocument.two},{JsonDocument.one}]"
                    "Documents not ordered correctly"
            }
        ]
        testList "jsonFirstByFields" [
            testTask "succeeds when a document is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonFirstByFields SqliteDb.TableName Any [ Field.Equal "Value" "another" ]
                Expect.equal json JsonDocument.two "The incorrect document was returned"
            }
            testTask "succeeds when multiple documents are found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonFirstByFields SqliteDb.TableName Any [ Field.Equal "Sub.Foo" "green" ]
                Expect.notEqual json "{}" "There should have been a document returned"
                verifyAny json [ JsonDocument.two; JsonDocument.four ]
            }
            testTask "succeeds when a document is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.jsonFirstByFields SqliteDb.TableName Any [ Field.Equal "Value" "absent" ]
                verifyNoDoc json
            }
        ]
        testList "jsonFirstByFieldsOrdered" [
            testTask "succeeds when sorting ascending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json =
                    conn.jsonFirstByFieldsOrdered
                        SqliteDb.TableName Any [ Field.Equal "Sub.Foo" "green" ] [ Field.Named "Sub.Bar" ]
                Expect.equal json JsonDocument.two "An incorrect document was returned"
            }
            testTask "succeeds when sorting descending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json =
                    conn.jsonFirstByFieldsOrdered
                        SqliteDb.TableName Any [ Field.Equal "Sub.Foo" "green" ] [ Field.Named "Sub.Bar DESC" ]
                Expect.equal json JsonDocument.four "An incorrect document was returned"
            }
        ]
        testList "writeJsonAll" [
            testTask "succeeds when there is data" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()

                do! conn.insert SqliteDb.TableName { Foo = "one"; Bar = "two" }
                do! conn.insert SqliteDb.TableName { Foo = "three"; Bar = "four" }
                do! conn.insert SqliteDb.TableName { Foo = "five"; Bar = "six" }

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonAll SqliteDb.TableName writer
                let json = streamText stream
                verifyBeginEnd json
                Expect.stringContains json """{"Foo":"one","Bar":"two"}"""    "The first document was not found"
                Expect.stringContains json """{"Foo":"three","Bar":"four"}""" "The second document was not found"
                Expect.stringContains json """{"Foo":"five","Bar":"six"}"""   "The third document was not found"
            }
            testTask "succeeds when there is no data" {
                use! db     = SqliteDb.BuildDb()
                use  conn   = Configuration.dbConn ()
                use  stream = new MemoryStream()
                use  writer = writeStream stream
                do! conn.writeJsonAll SqliteDb.TableName writer
                verifyEmpty (streamText stream)
            }
        ]
        testList "writeJsonAllOrdered" [
            testTask "succeeds when ordering numerically" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonAllOrdered SqliteDb.TableName writer [ Field.Named "n:NumValue" ]
                Expect.equal
                    (streamText stream)
                    $"[{JsonDocument.one},{JsonDocument.three},{JsonDocument.two},{JsonDocument.four},{JsonDocument.five}]"
                    "The documents were not ordered correctly"
            }
            testTask "succeeds when ordering numerically descending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonAllOrdered SqliteDb.TableName writer [ Field.Named "n:NumValue DESC" ]
                Expect.equal
                    (streamText stream)
                    $"[{JsonDocument.five},{JsonDocument.four},{JsonDocument.two},{JsonDocument.three},{JsonDocument.one}]"
                    "The documents were not ordered correctly"
            }
            testTask "succeeds when ordering alphabetically" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonAllOrdered SqliteDb.TableName writer [ Field.Named "Id DESC" ]
                Expect.equal
                    (streamText stream)
                    $"[{JsonDocument.two},{JsonDocument.three},{JsonDocument.one},{JsonDocument.four},{JsonDocument.five}]"
                    "The documents were not ordered correctly"
            }
        ]
        testList "writeJsonById" [
            testTask "succeeds when a document is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonById SqliteDb.TableName writer "two"
                Expect.equal (streamText stream) JsonDocument.two "The incorrect document was returned"
            }
            testTask "succeeds when a document is not found" {
                use! db = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonById SqliteDb.TableName writer "three hundred eighty-seven"
                verifyNoDoc (streamText stream)
            }
        ]
        testList "writeJsonByFields" [
            testTask "succeeds when documents are found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonByFields SqliteDb.TableName writer Any [ Field.Greater "NumValue" 15 ]
                let json = streamText stream
                verifyBeginEnd json
                Expect.stringContains json JsonDocument.four "Document `four` should have been returned"
                Expect.stringContains json JsonDocument.five "Document `five` should have been returned"
            }
            testTask "succeeds when documents are found using IN with numeric field" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonByFields SqliteDb.TableName writer All [ Field.In "NumValue" [ 2; 4; 6; 8 ] ]
                Expect.equal (streamText stream) $"[{JsonDocument.three}]" "There should have been one document returned"
            }
            testTask "succeeds when documents are not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonByFields SqliteDb.TableName writer Any [ Field.Greater "NumValue" 100 ]
                verifyEmpty (streamText stream)
            }
            testTask "succeeds for InArray when matching documents exist" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! conn.ensureTable SqliteDb.TableName
                for doc in ArrayDocument.TestDocuments do do! conn.insert SqliteDb.TableName doc

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonByFields
                        SqliteDb.TableName writer All [ Field.InArray "Values" SqliteDb.TableName [ "c" ] ]
                let json = streamText stream
                verifyBeginEnd json
                Expect.stringContains
                    json """{"Id":"first","Values":["a","b","c"]}""" "Document `first` should have been returned"
                Expect.stringContains
                    json """{"Id":"second","Values":["c","d","e"]}""" "Document `second` should have been returned"
            }
            testTask "succeeds for InArray when no matching documents exist" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! conn.ensureTable SqliteDb.TableName
                for doc in ArrayDocument.TestDocuments do do! conn.insert SqliteDb.TableName doc

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonByFields
                        SqliteDb.TableName writer All [ Field.InArray "Values" SqliteDb.TableName [ "j" ] ]
                verifyEmpty (streamText stream)
            }
        ]
        testList "writeJsonByFieldsOrdered" [
            testTask "succeeds when sorting ascending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonByFieldsOrdered
                        SqliteDb.TableName writer Any [ Field.Greater "NumValue" 15 ] [ Field.Named "Id" ]
                Expect.equal
                    (streamText stream) $"[{JsonDocument.five},{JsonDocument.four}]" "Incorrect documents were returned"
            }
            testTask "succeeds when sorting descending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonByFieldsOrdered
                        SqliteDb.TableName writer Any [ Field.Greater "NumValue" 15 ] [ Field.Named "Id DESC" ]
                Expect.equal
                    (streamText stream) $"[{JsonDocument.four},{JsonDocument.five}]" "Incorrect documents were returned"
            }
            testTask "succeeds when sorting case-sensitively" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonByFieldsOrdered
                        SqliteDb.TableName writer All [ Field.LessOrEqual "NumValue" 10 ] [ Field.Named "Value" ]
                Expect.equal
                    (streamText stream)
                    $"[{JsonDocument.three},{JsonDocument.one},{JsonDocument.two}]"
                    "Documents not ordered correctly"
            }
            testTask "succeeds when sorting case-insensitively" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonByFieldsOrdered
                        SqliteDb.TableName writer All [ Field.LessOrEqual "NumValue" 10 ] [ Field.Named "i:Value" ]
                Expect.equal
                    (streamText stream)
                    $"[{JsonDocument.three},{JsonDocument.two},{JsonDocument.one}]"
                    "Documents not ordered correctly"
            }
        ]
        testList "writeJsonFirstByFields" [
            testTask "succeeds when a document is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonFirstByFields SqliteDb.TableName writer Any [ Field.Equal "Value" "another" ]
                Expect.equal (streamText stream) JsonDocument.two "The incorrect document was returned"
            }
            testTask "succeeds when multiple documents are found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonFirstByFields SqliteDb.TableName writer Any [ Field.Equal "Sub.Foo" "green" ]
                let json = streamText stream
                Expect.notEqual json "{}" "There should have been a document returned"
                verifyAny json [ JsonDocument.two; JsonDocument.four ]
            }
            testTask "succeeds when a document is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonFirstByFields SqliteDb.TableName writer Any [ Field.Equal "Value" "absent" ]
                verifyNoDoc (streamText stream)
            }
        ]
        testList "writeJsonFirstByFieldsOrdered" [
            testTask "succeeds when sorting ascending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonFirstByFieldsOrdered
                        SqliteDb.TableName writer Any [ Field.Equal "Sub.Foo" "green" ] [ Field.Named "Sub.Bar" ]
                Expect.equal (streamText stream) JsonDocument.two "An incorrect document was returned"
            }
            testTask "succeeds when sorting descending" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeJsonFirstByFieldsOrdered
                        SqliteDb.TableName writer Any [ Field.Equal "Sub.Foo" "green" ] [ Field.Named "Sub.Bar DESC" ]
                Expect.equal (streamText stream) JsonDocument.four "An incorrect document was returned"
            }
        ]
        testList "updateById" [
            testTask "succeeds when a document is updated" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let testDoc = { emptyDoc with Id = "one"; Sub = Some { Foo = "blue"; Bar = "red" } }
                do! conn.updateById SqliteDb.TableName "one" testDoc
                let! after = conn.findById<string, JsonDocument> SqliteDb.TableName "one"
                if Option.isNone after then
                    Expect.isTrue false "There should have been a document returned post-update"
                Expect.equal after.Value testDoc "The updated document is not correct"
            }
            testTask "succeeds when no document is updated" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()

                let! before = conn.findAll<JsonDocument> SqliteDb.TableName
                Expect.isEmpty before "There should have been no documents returned"

                // This not raising an exception is the test
                do! conn.updateById
                        SqliteDb.TableName
                        "test"
                        { emptyDoc with Id = "x"; Sub = Some { Foo = "blue"; Bar = "red" } }
            }
        ]
        testList "updateByFunc" [
            testTask "succeeds when a document is updated" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.updateByFunc
                        SqliteDb.TableName _.Id { Id = "one"; Value = "le un"; NumValue = 1; Sub = None }
                let! after = conn.findById<string, JsonDocument> SqliteDb.TableName "one"
                if Option.isNone after then
                    Expect.isTrue false "There should have been a document returned post-update"
                Expect.equal
                    after.Value
                    { Id = "one"; Value = "le un"; NumValue = 1; Sub = None }
                    "The updated document is not correct"
            }
            testTask "succeeds when no document is updated" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()

                let! before = conn.findAll<JsonDocument> SqliteDb.TableName
                Expect.isEmpty before "There should have been no documents returned"

                // This not raising an exception is the test
                do! conn.updateByFunc
                        SqliteDb.TableName _.Id { Id = "one"; Value = "le un"; NumValue = 1; Sub = None }
            }
        ]
        testList "patchById" [
            testTask "succeeds when a document is updated" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.patchById SqliteDb.TableName "one" {| NumValue = 44 |}
                let! after = conn.findById<string, JsonDocument> SqliteDb.TableName "one"
                if Option.isNone after then
                    Expect.isTrue false "There should have been a document returned post-update"
                Expect.equal after.Value.NumValue 44 "The updated document is not correct"
            }
            testTask "succeeds when no document is updated" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()

                let! before = conn.findAll<SubDocument> SqliteDb.TableName
                Expect.isEmpty before "There should have been no documents returned"

                // This not raising an exception is the test
                do! conn.patchById SqliteDb.TableName "test" {| Foo = "green" |}
            }
        ]
        testList "patchByFields" [
            testTask "succeeds when a document is updated" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.patchByFields SqliteDb.TableName Any [ Field.Equal "Value" "purple" ] {| NumValue = 77 |}
                let! after = conn.countByFields SqliteDb.TableName Any [ Field.Equal "NumValue" 77 ]
                Expect.equal after 2L "There should have been 2 documents returned"
            }
            testTask "succeeds when no document is updated" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()

                let! before = conn.findAll<SubDocument> SqliteDb.TableName
                Expect.isEmpty before "There should have been no documents returned"

                // This not raising an exception is the test
                do! conn.patchByFields SqliteDb.TableName Any [ Field.Equal "Value" "burgundy" ] {| Foo = "green" |}
            }
        ]
        testList "removeFieldsById" [
            testTask "succeeds when fields are removed" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.removeFieldsById SqliteDb.TableName "two" [ "Sub"; "Value" ]
                try
                    let! _ = conn.findById<string, JsonDocument> SqliteDb.TableName "two"
                    Expect.isTrue false "The updated document should have failed to parse"
                with
                | :? JsonException -> ()
                | exn as ex -> Expect.isTrue false $"Threw {ex.GetType().Name} ({ex.Message})"
            }
            testTask "succeeds when a field is not removed" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                // This not raising an exception is the test
                do! conn.removeFieldsById SqliteDb.TableName "two" [ "AFieldThatIsNotThere" ]
            }
            testTask "succeeds when no document is matched" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()

                // This not raising an exception is the test
                do! conn.removeFieldsById SqliteDb.TableName "two" [ "Value" ]
            }
        ]
        testList "removeFieldByFields" [
            testTask "succeeds when a field is removed" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.removeFieldsByFields SqliteDb.TableName Any [ Field.Equal "NumValue" 17 ] [ "Sub" ]
                try
                    let! _ = conn.findById<string, JsonDocument> SqliteDb.TableName "four"
                    Expect.isTrue false "The updated document should have failed to parse"
                with
                | :? JsonException -> ()
                | exn as ex -> Expect.isTrue false $"Threw {ex.GetType().Name} ({ex.Message})"
            }
            testTask "succeeds when a field is not removed" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                // This not raising an exception is the test
                do! conn.removeFieldsByFields SqliteDb.TableName Any [ Field.Equal "NumValue" 17 ] [ "Nothing" ]
            }
            testTask "succeeds when no document is matched" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()

                // This not raising an exception is the test
                do! conn.removeFieldsByFields
                        SqliteDb.TableName Any [ Field.NotEqual "Abracadabra" "apple" ] [ "Value" ]
            }
        ]
        testList "deleteById" [
            testTask "succeeds when a document is deleted" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.deleteById SqliteDb.TableName "four"
                let! remaining = conn.countAll SqliteDb.TableName
                Expect.equal remaining 4L "There should have been 4 documents remaining"
            }
            testTask "succeeds when a document is not deleted" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.deleteById SqliteDb.TableName "thirty"
                let! remaining = conn.countAll SqliteDb.TableName
                Expect.equal remaining 5L "There should have been 5 documents remaining"
            }
        ]
        testList "deleteByFields" [
            testTask "succeeds when documents are deleted" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.deleteByFields SqliteDb.TableName Any [ Field.NotEqual "Value" "purple" ]
                let! remaining = conn.countAll SqliteDb.TableName
                Expect.equal remaining 2L "There should have been 2 documents remaining"
            }
            testTask "succeeds when documents are not deleted" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.deleteByFields SqliteDb.TableName Any [ Field.Equal "Value" "crimson" ]
                let! remaining = conn.countAll SqliteDb.TableName
                Expect.equal remaining 5L "There should have been 5 documents remaining"
            }
        ]
        testList "customList" [
            testTask "succeeds when data is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! docs = conn.customList (Query.find SqliteDb.TableName) [] fromData<JsonDocument>
                Expect.hasLength docs 5 "There should have been 5 documents returned"
            }
            testTask "succeeds when data is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! docs =
                    conn.customList
                        $"SELECT data FROM {SqliteDb.TableName} WHERE data ->> 'NumValue' > @value"
                        [ SqliteParameter("@value", 100) ]
                        fromData<JsonDocument>
                Expect.isEmpty docs "There should have been no documents returned"
            }
        ]
        testList "customJsonArray" [
            testTask "succeeds when data is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json = conn.customJsonArray (Query.find SqliteDb.TableName) [] jsonFromData
                Expect.stringStarts json "[" "The JSON array should have started with `[`"
                Expect.stringContains json JsonDocument.one   "Document ID `one` should have been found"
                Expect.stringContains json JsonDocument.two   "Document ID `two` should have been found"
                Expect.stringContains json JsonDocument.three "Document ID `three` should have been found"
                Expect.stringContains json JsonDocument.four  "Document ID `four` should have been found"
                Expect.stringContains json JsonDocument.five  "Document ID `five` should have been found"
                Expect.stringEnds json "]" "The JSON array should have ended with `[`"
            }
            testTask "succeeds when data is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! docs =
                    conn.customJsonArray
                        $"SELECT data FROM {SqliteDb.TableName} WHERE data ->> 'NumValue' > @value"
                        [ SqliteParameter("@value", 100) ]
                        jsonFromData
                Expect.equal docs "[]" "There should have been no documents returned"
            }
        ]
        testList "writeCustomJsonArray" [
            testTask "succeeds when data is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeCustomJsonArray (Query.find SqliteDb.TableName) [] writer jsonFromData

                let json = streamText stream
                Expect.stringStarts json "[" "The JSON array should have started with `[`"
                Expect.stringContains json JsonDocument.one   "Document ID `one` should have been found"
                Expect.stringContains json JsonDocument.two   "Document ID `two` should have been found"
                Expect.stringContains json JsonDocument.three "Document ID `three` should have been found"
                Expect.stringContains json JsonDocument.four  "Document ID `four` should have been found"
                Expect.stringContains json JsonDocument.five  "Document ID `five` should have been found"
                Expect.stringEnds json "]" "The JSON array should have ended with `[`"
            }
            testTask "succeeds when data is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                use stream = new MemoryStream()
                use writer = writeStream stream
                do! conn.writeCustomJsonArray
                        $"SELECT data FROM {SqliteDb.TableName} WHERE data ->> 'NumValue' > @value"
                        [ SqliteParameter("@value", 100) ]
                        writer
                        jsonFromData

                Expect.equal (streamText stream) "[]" "There should have been no documents returned"
            }
        ]
        testList "customSingle" [
            testTask "succeeds when a row is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! doc =
                    conn.customSingle
                        $"SELECT data FROM {SqliteDb.TableName} WHERE data ->> 'Id' = @id"
                        [ SqliteParameter("@id", "one") ]
                        fromData<JsonDocument>
                Expect.isSome doc "There should have been a document returned"
                Expect.equal doc.Value.Id "one" "The incorrect document was returned"
            }
            testTask "succeeds when a row is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! doc =
                    conn.customSingle
                        $"SELECT data FROM {SqliteDb.TableName} WHERE data ->> 'Id' = @id"
                        [ SqliteParameter("@id", "eighty") ]
                        fromData<JsonDocument>
                Expect.isNone doc "There should not have been a document returned"
            }
        ]
        testList "customJsonSingle" [
            testTask "succeeds when a row is found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json =
                    conn.customJsonSingle
                        $"SELECT data FROM {SqliteDb.TableName} WHERE data ->> 'Id' = @id"
                        [ SqliteParameter("@id", "one") ]
                        jsonFromData
                Expect.equal json JsonDocument.one "The JSON document is incorrect"
            }
            testTask "succeeds when a row is not found" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                let! json =
                    conn.customJsonSingle
                        $"SELECT data FROM {SqliteDb.TableName} WHERE data ->> 'Id' = @id"
                        [ SqliteParameter("@id", "eighty") ]
                        jsonFromData
                Expect.equal json "{}" "There should not have been a document returned"
            }
        ]
        testList "customNonQuery" [
            testTask "succeeds when operating on data" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.customNonQuery $"DELETE FROM {SqliteDb.TableName}" []

                let! remaining = conn.countAll SqliteDb.TableName
                Expect.equal remaining 0L "There should be no documents remaining in the table"
            }
            testTask "succeeds when no data matches where clause" {
                use! db   = SqliteDb.BuildDb()
                use  conn = Configuration.dbConn ()
                do! loadDocs conn

                do! conn.customNonQuery
                        $"DELETE FROM {SqliteDb.TableName} WHERE data ->> 'NumValue' > @value"
                        [ SqliteParameter("@value", 100) ]

                let! remaining = conn.countAll SqliteDb.TableName
                Expect.equal remaining 5L "There should be 5 documents remaining in the table"
            }
        ]
        testTask "customScalar succeeds" {
            use! db   = SqliteDb.BuildDb()
            use  conn = Configuration.dbConn ()

            let! nbr = conn.customScalar "SELECT 5 AS test_value" [] _.GetInt32(0)
            Expect.equal nbr 5 "The query should have returned the number 5"
        }
        test "clean up database" {
            Configuration.useConnectionString "data source=:memory:"
        }
    ]
    |> testSequenced
